"use strict";
Object.defineProperty(exports, "__esModule", {
    value: true
});
function _export(target, all) {
    for(var name in all)Object.defineProperty(target, name, {
        enumerable: true,
        get: all[name]
    });
}
_export(exports, {
    ArkTypeError: ()=>ArkTypeError,
    Problem: ()=>Problem,
    Problems: ()=>Problems,
    describeDomains: ()=>describeDomains,
    describeBranches: ()=>describeBranches,
    DataWrapper: ()=>DataWrapper,
    defaultProblemWriters: ()=>defaultProblemWriters
});
const _scannerTs = require("../parse/string/shift/scanner.js");
const _domainsTs = require("../utils/domains.js");
const _numericLiteralsTs = require("../utils/numericLiterals.js");
const _objectKindsTs = require("../utils/objectKinds.js");
const _pathsTs = require("../utils/paths.js");
const _serializeTs = require("../utils/serialize.js");
var _definitions_code, _definitions_code1, _clausesByPath, _pathKey;
class ArkTypeError extends TypeError {
    constructor(problems){
        super(problems.summary);
        this.cause = problems;
    }
}
class Problem {
    toString() {
        return this.message;
    }
    get message() {
        return this.writers.addContext(this.reason, this.path);
    }
    get reason() {
        return this.writers.writeReason(this.mustBe, new DataWrapper(this.data));
    }
    get mustBe() {
        return this.writers.mustBe(this.source);
    }
    constructor(code, path, data, source, writers){
        this.code = code;
        this.path = path;
        this.data = data;
        this.source = source;
        this.writers = writers;
        if (this.code === "multi") {
            this.parts = this.source;
        }
    }
}
class ProblemArray extends Array {
    add(code, data, source) {
        const problem = new Problem(code, // copy the path to avoid future mutations affecting it
        _pathsTs.Path.from(this.state.path), data, source, this.state.getConfigForProblemCode(code));
        const pathKey = `${this.state.path}`;
        const existing = this.byPath[pathKey];
        if (existing) {
            if (existing.parts) {
                existing.parts.push(problem);
            } else {
                const problemIntersection = new Problem("multi", existing.path, data, [
                    existing,
                    problem
                ], this.state.getConfigForProblemCode("multi"));
                const existingIndex = this.indexOf(existing);
                // If existing is found (which it always should be unless this was externally mutated),
                // replace it with the new problem intersection. In case it isn't for whatever reason,
                // just append the intersection.
                this[existingIndex === -1 ? this.length : existingIndex] = problemIntersection;
                this.byPath[pathKey] = problemIntersection;
            }
        } else {
            this.byPath[pathKey] = problem;
            this.push(problem);
        }
    }
    get summary() {
        if (this.length === 1) {
            return this[0].message;
        }
        return this.join("\n");
    }
    toString() {
        return this.summary;
    }
    throw() {
        throw new ArkTypeError(this);
    }
    constructor(state){
        super();
        this.state = state;
        this.byPath = {};
    }
}
const Problems = ProblemArray;
const capitalize = (s)=>s[0].toUpperCase() + s.slice(1);
const describeDomains = (domains)=>{
    if (domains.length === 1) {
        return _domainsTs.domainDescriptions[domains[0]];
    }
    if (domains.length === 0) {
        return "never";
    }
    return describeBranches(domains.map((objectKind)=>_domainsTs.domainDescriptions[objectKind]));
};
const describeBranches = (descriptions)=>{
    let description = "";
    for(let i = 0; i < descriptions.length - 1; i++){
        description += descriptions[i];
        if (i < descriptions.length - 2) {
            description += ", ";
        }
    }
    description += ` or ${descriptions[descriptions.length - 1]}`;
    return description;
};
class DataWrapper {
    toString() {
        return (0, _serializeTs.stringify)(this.value);
    }
    get domain() {
        return (0, _domainsTs.domainOf)(this.value);
    }
    get objectKind() {
        return (0, _objectKindsTs.objectKindOf)(this.value) ?? "non-object";
    }
    get size() {
        return typeof this.value === "string" || Array.isArray(this.value) ? this.value.length : typeof this.value === "number" ? this.value : 0;
    }
    get units() {
        return typeof this.value === "string" ? "characters" : Array.isArray(this.value) ? "items" : "";
    }
    get className() {
        return Object(this.value).constructor.name;
    }
    constructor(value){
        this.value = value;
    }
}
const writeDefaultReason = (mustBe, was)=>`must be ${mustBe}${was && ` (was ${was})`}`;
const writeDefaultProblemMessage = (reason, path)=>path.length === 0 ? capitalize(reason) : path.length === 1 && (0, _numericLiteralsTs.isWellFormedInteger)(path[0]) ? `Item at index ${path[0]} ${reason}` : `${path} ${reason}`;
const compileDefaultProblemWriters = (definitions)=>{
    let code;
    for(code in definitions){
        (_definitions_code = definitions[code]).writeReason ?? (_definitions_code.writeReason = writeDefaultReason);
        (_definitions_code1 = definitions[code]).addContext ?? (_definitions_code1.addContext = writeDefaultProblemMessage);
    }
    return definitions;
};
const defaultProblemWriters = compileDefaultProblemWriters({
    divisor: {
        mustBe: (divisor)=>divisor === 1 ? `an integer` : `a multiple of ${divisor}`
    },
    class: {
        mustBe: (constructor)=>`an instance of ${constructor.name}`,
        writeReason: (mustBe, data)=>writeDefaultReason(mustBe, data.className)
    },
    domain: {
        mustBe: (domain)=>_domainsTs.domainDescriptions[domain],
        writeReason: (mustBe, data)=>writeDefaultReason(mustBe, data.domain)
    },
    domainBranches: {
        mustBe: (domains)=>describeDomains(domains),
        writeReason: (mustBe, data)=>writeDefaultReason(mustBe, data.domain)
    },
    objectKind: {
        mustBe: (kind)=>_objectKindsTs.objectKindDescriptions[kind],
        writeReason: (mustBe, data)=>writeDefaultReason(mustBe, data.objectKind)
    },
    missing: {
        mustBe: ()=>"defined",
        writeReason: (mustBe)=>writeDefaultReason(mustBe, "")
    },
    bound: {
        mustBe: (bound)=>`${_scannerTs.Scanner.comparatorDescriptions[bound.comparator]} ${bound.limit}${bound.units ? ` ${bound.units}` : ""}`,
        writeReason: (mustBe, data)=>writeDefaultReason(mustBe, `${data.size}`)
    },
    regex: {
        mustBe: (expression)=>expression[0] === "/" ? `a string matching ${expression}` : `a valid ${expression}`
    },
    value: {
        mustBe: _serializeTs.stringify
    },
    valueBranches: {
        mustBe: (values)=>describeBranches(values.map(_serializeTs.stringify))
    },
    branches: {
        mustBe: (branchProblems)=>{
            const clausesByPath = {};
            for (const problem of branchProblems){
                const pathKey = `${problem.path}`;
                (_clausesByPath = clausesByPath)[_pathKey = pathKey] ?? (_clausesByPath[_pathKey] = []);
                clausesByPath[pathKey].push(problem.mustBe);
            }
            return describeBranches(Object.entries(clausesByPath).map(([path, clauses])=>`${path} must be ${clauses.length === 1 ? clauses[0] : describeBranches(clauses)}`));
        },
        writeReason: (mustBe, data)=>`${mustBe} (was ${data})`,
        addContext: (reason, path)=>path.length ? `At ${path}, ${reason}` : reason
    },
    multi: {
        mustBe: (problems)=>"• " + problems.map((_)=>_.mustBe).join("\n• "),
        writeReason: (mustBe, data)=>`${data} must be...\n${mustBe}`,
        addContext: (reason, path)=>path.length ? `At ${path}, ${reason}` : reason
    }
});
